/**
 * Copyright 2009 Pierrick HYMBERT.
 * Les sources qui constituent ce projet, HyPik-Soma-Cube, 
 * de mme que la documentation associe sont la proprit
 * de leur auteur.
 * Je donne mon accord au site developpez.com pour l'utili-
 * sation de tout ou partie des sources et de la documenta-
 * tion de ce projet dans les pages developpez.com
 */
package fr.hypik.developpez.defis.soma.core.model;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;

import javax.vecmath.Matrix3d;
import javax.vecmath.Point3d;

import org.apache.log4j.Logger;

/**
 * Un lment de base reprsentant un point dans l'espace P(x,y,z)
 * @author Pierrick HYMBERT
 */
public class Cube {
	/**
	 * Le logger de la classe
	 */
	private static final Logger logger = Logger.getLogger(Cube.class);
	
	/**
	 * Le support pour les listeners
	 */
	private PropertyChangeSupport propertyChangeSupport = new PropertyChangeSupport(this);
	
	/**
	 * Le cube centre
	 */
	public static final Cube ORIGINE = new Cube(0,0,0);
	
	/**
	 * L'abscisse du cube
	 */
	private int x;
	
	/**
	 * L'ordonne du cube
	 */
	private int y;
	
	/**
	 * La profondeur du cube
	 */
	private int z;
	
	/**
	 * Le constructeur par dfaut
	 */
	public Cube() {
	}
	
	/**
	 * Le constructeur complet
	 * @param x L'abscisse du cube
	 * @param y L'ordonne du cube
	 * @param z La profondeur du cube
	 */
	public Cube(int x, int y, int z) {
		super();
		this.x = x;
		this.y = y;
		this.z = z;
		logger.debug("Construction du cube: " + this);
	}
	

	/**
	 * @return L'abscisse du cube
	 */
	public int getX() {
		return x;
	}

	/**
	 * @return L'ordonne du cube
	 */
	public int getY() {
		return y;
	}
	
	/**
	 * @return La profondeur du cube
	 */
	public int getZ() {
		return z;
	}

	/**
	 * @return La valeur sur l'axe
	 */
	public int get(Axe axe) {
		int coordonnee = 0;
		switch (axe) {
		case Abscisse:
			coordonnee = x;
			break;
		case Ordonnee:
			coordonnee = y;
			break;
		case Profondeur:
			coordonnee = z;
			break;
		}
		return coordonnee;
	}
	/**
	 * Retourne la distance entre deux cube sur un plan
	 * @param cube Le cube  comparer
	 * @param axe1 L'axe de la distance  mesurer
	 * @param axe2 L'axe de la distance  mesurer
	 * @return La distance
	 */
	public int distance(Cube cube, Axe axe1, Axe axe2)
	{
		return (int) Math.sqrt( 	Math.pow(get(axe1) - cube.get(axe1), 2)
				+ Math.pow(get(axe2) - cube.get(axe2), 2));
	}
	
	/**
	 * Retourne la distance entre deux cube sur un axe
	 * @param cube Le cube  comparer
	 * @param axe L'axe de la distance  mesurer
	 * @return La distance
	 */
	public int distance(Cube cube, Axe axe)
	{
		int distance = 0;
		switch (axe) {
		case Abscisse:
			distance = x - cube.x;
			break;
		case Ordonnee:
			distance = y - cube.y;
			break;
		case Profondeur:
			distance = z - cube.z;
			break;
		}
		return distance;
	}
	
	/**
	 * Retourne la distance entre deux cube sur un axe
	 * @param cube Le cube  comparer
	 * @return La distance
	 */
	public double distance(Cube cube)
	{
		return Math.sqrt( 	Math.pow(x - cube.x, 2)
							+ Math.pow(y - cube.y, 2)
							+ Math.pow(z - cube.z, 2));
	}
	
	/**
	 * Translate le cube un angle et une distance
	 * @param axeL'axe de la translation
	 * @param distance L'axe de la rotation
	 * @return This pour le chainage des methodes
	 */
	public Cube translation(Axe axe, int distance)
	{
		String toString = null;
		if( logger.isDebugEnabled() )
			toString = toString();
		switch (axe) {
		case Abscisse:
			this.x += distance;
			break;
		case Ordonnee:
			this.y += distance;
			break;
		case Profondeur:
			this.z += distance;
			break;
		}
		if( logger.isDebugEnabled() )
			logger.debug("Translation du cube " + toString + " sur l'axe " + axe + " d'un distance " + distance + " = " + this);
		
		// Avertie les listeners
		propertyChangeSupport.firePropertyChange(null, null, this);
		
		return this;
	}

	/**
	 * Fait tourner le cube selon un cube et un angle autour de l'axe
	 * @param axe L'axe de la rotation
	 * @param angle L'angle de la rotation (Un entier qui sera multipli par PI/2=90, et oui on gre que des carrs)
	 * @return This pour le chainage des methodes
	 */
	public Cube rotation(Axe axe, int angle)
	{
		String toString = null;
		if( logger.isDebugEnabled() )
			toString = toString();

		Matrix3d matrix3d = new Matrix3d();
		Point3d point3d = new Point3d(x, y, z);
		double angleRad = Math.PI / 2 * angle;
		switch (axe) {
		case Abscisse:
			matrix3d.rotX(angleRad);
			matrix3d.transform(point3d);
			break;
		case Ordonnee:
			matrix3d.rotY(angleRad);
			matrix3d.transform(point3d);
			break;
		case Profondeur:
			matrix3d.rotZ(angleRad);
			matrix3d.transform(point3d);
			break;
		}
		
		this.x = (int) Math.round(point3d.x);
		this.y = (int) Math.round(point3d.y);
		this.z = (int) Math.round(point3d.z);
		
		if( logger.isDebugEnabled() )
			logger.debug("Rotation du cube " + toString + " sur l'axe " + axe + " d'un angle " + angle + " * PI/2 = " + this);
		
		// Avertie les listeners
		propertyChangeSupport.firePropertyChange(null, null, this);
		
		return this;
	}
	
	/*
	 * (non-Javadoc)
	 * @see java.lang.Object#hashCode()
	 */
	@Override
	public int hashCode() {
		return 23 * x + 29 * y + 31 * z;
	}
	
	/*
	 * (non-Javadoc)
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(Object obj) {
		if( obj instanceof Cube )
		{
			Cube cube = (Cube) obj;
			return 		x == cube.x 
					&&	y == cube.y
					&&	z == cube.z;
		}
		return super.equals(obj);
	}
	
	/*
	 * (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	@Override
	public String toString() {
		return "(" + x + "," + y + "," + z + ")";
	}
	
	/*
	 * (non-Javadoc)
	 * @see java.lang.Object#clone()
	 */
	@Override
	public Object clone() throws CloneNotSupportedException {
		return new Cube(x, y, z);
	}
	
	/**
	 * Ajoute un listener
	 * @param propertyChangeListener le listener  ajouter
	 */
	public void addPropertyChangeListener(PropertyChangeListener propertyChangeListener)
	{
		propertyChangeSupport.addPropertyChangeListener(propertyChangeListener);
	}
	
	/**
	 * Supprime un listener
	 * @param propertyChangeListener le listener  ajouter
	 */
	public void removePropertyChangeListener(PropertyChangeListener propertyChangeListener)
	{
		propertyChangeSupport.removePropertyChangeListener(propertyChangeListener);
	}
}
