/**
 * Copyright 2009 Pierrick HYMBERT.
 * Les sources qui constituent ce projet, HyPik-Soma-Cube, 
 * de mme que la documentation associe sont la proprit
 * de leur auteur.
 * Je donne mon accord au site developpez.com pour l'utili-
 * sation de tout ou partie des sources et de la documenta-
 * tion de ce projet dans les pages developpez.com
 */
package fr.hypik.developpez.defis.soma.core.model;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.util.ArrayList;
import java.util.List;

import org.apache.log4j.Logger;

/**
 * La classe reprsentant une des sept pices du cube de soma
 * 
 * @author Pierrick HYMBERT
 *
 */
public class Piece {

	/**
	 * Serialization classe unique identificateur
	 */
	private static final long serialVersionUID = 1L;
	
	/**
	 * La piece centre
	 */
	public static final Piece ORIGINE = new Piece("Origine", new Cube[]{Cube.ORIGINE});
	
	/**
	 * Le logger de la classe
	 */
	private static final Logger logger = Logger.getLogger(Piece.class);
	
	/**
	 * Le nom de la piece
	 */
	private String nom;
	
	/**
	 * Les cubes de la pice
	 */
	private Cube[] cubes;
	
	/**
	 * Le support pour les listeners
	 */
	private PropertyChangeSupport propertyChangeSupport = new PropertyChangeSupport(this);
	
	/**
	 * Construit une pice selon 
	 */
	public Piece(String nom, Cube[] cubes) {
		this.nom = nom;
		this.cubes = cubes;
	}

	/**
	 * Fait tourner la pice selon un cube et un angle autour de l'axe
	 * @param centre Le centre de la rotation
	 * @param axe L'axe de la rotation
	 * @param angle L'angle de la rotation (Un entier qui sera multipli par PI/2=90, et oui on gre que des carrs)
	 */
	public void rotation(Axe axe, int angle)
	{
		String toString = null;
		if( logger.isDebugEnabled() )
			toString = toString();
		for(Cube cube : cubes)
		{
			cube.rotation(axe, angle);
		}
		if( logger.isDebugEnabled() )
			logger.debug("Rotation de la piece " + toString + " sur l'axe " + axe + " d'un angle " + angle + " * PI/2 = " + this);
		// Avertie les listeners
		propertyChangeSupport.firePropertyChange(null, null, this);
	}
	
	/**
	 * Verifie si la pice est compltement incluse dans celle ci
	 * @param pieceIncluse La pice qui est incluse
	 * @return True si pieceIncluse est incluse dans cette piece, false sinon
	 */
	public boolean include(Piece pieceIncluse)
	{
		for(Cube cubeInclu : pieceIncluse.cubes)
		{
			boolean include = false;
			for(Cube cube : cubes)
			{
				if( cubeInclu.equals(cube) )
				{
					include = true;
					break;
				}
			}
			if( !include )
				return false;
		}
		
		return true;
	}
	
	/**
	 * Verifie si la pice est en collision avec celle ci
	 * @param pieceIncluse La pice qui est incluse
	 * @return True si pieceIncluse est incluse dans cette piece, false sinon
	 */
	public boolean collide(Piece piece)
	{
		for(Cube cubeCollide : piece.cubes)
		{
			for(Cube cube : cubes)
			{
				if( cubeCollide.equals(cube) )
				{
					return true;
				}
			}
		}
		
		return false;
	}
	
	/**
	 * Retourne la plus courte distance entre deux pices cube  cube
	 * @param piece La pice distante
	 * @return True si pieceIncluse est incluse dans cette piece, false sinon
	 */
	public double distance(Piece piece)
	{
		double distance = Double.MAX_VALUE;
		for(Cube cubePiece : piece.cubes)
		{
			for(Cube cube : cubes)
			{
				distance = Math.min(distance, cubePiece.distance(cube));
			}
		}
		
		return distance;
	}
	
	/**
	 * Retourne la plus courte distance entre deux piece sur un axe
	 * @param piece La piece  comparer
	 * @param axe L'axe de la distance  mesurer
	 * @return La distance
	 */
	public int distanceMin(Piece piece, Axe axe)
	{
		int distance = Integer.MAX_VALUE;
		for(Cube cubePiece : piece.cubes)
		{
			for(Cube cube : cubes)
			{
				distance = Math.min(distance, cube.distance(cubePiece, axe));
			}
		}
		
		return distance;
	}
	
	/**
	 * Retourne la plus courte distance entre deux piece sur un axe
	 * @param piece La piece  comparer
	 * @param axe L'axe de la distance  mesurer
	 * @return La distance
	 */
	public int distanceMax(Piece piece, Axe axe)
	{
		int distance = Integer.MIN_VALUE;
		for(Cube cubePiece : piece.cubes)
		{
			for(Cube cube : cubes)
			{
				distance = Math.max(distance, cube.distance(cubePiece, axe));
			}
		}
		
		return distance;
	}
	
	/**
	 * Retourne la plus courte distance entre deux piece sur un axe
	 * @param piece La piece  comparer
	 * @param axe L'axe de la distance  mesurer
	 * @return La distance
	 */
	public int distanceMoy(Piece piece, Axe axe)
	{
		int distance = 0;
		for(Cube cubePiece : piece.cubes)
		{
			for(Cube cube : cubes)
			{
				distance += cube.distance(cubePiece, axe);
			}
		}
		
		return piece.cubes.length > 0 && cubes.length > 0 ? (int)Math.round((double)distance / (double)( piece.cubes.length * cubes.length)) : 0;
	}
	
	/**
	 * Translate la pice un angle et une distance
	 * @param axe L'axe de la translation
	 * @param distance L'axe de la rotation
	 */
	public void translation(Axe axe, int distance)
	{
		String toString = null;
		if( logger.isDebugEnabled() )
			toString = toString();
		for(Cube cube : cubes)
		{
			cube.translation(axe, distance);
		}
		if( logger.isDebugEnabled() )
			logger.debug("Translation de la piece " + toString + " sur l'axe " + axe + " d'un distance " + distance + " = " + this);
		// Avertie les listeners
		propertyChangeSupport.firePropertyChange(null, null, this);
	}
	
	@Override
	public int hashCode() {
		int hash = 19;
		for(Cube cube : cubes)
		{
			hash += cube.hashCode();
		}
		return hash;
	}
	
	@Override
	public boolean equals(Object obj) {
		if( obj instanceof Piece )
		{
			Piece piece = (Piece) obj;
			List<Cube> cubeToFind = new ArrayList<Cube>();
			for(Cube cube : cubes)
				cubeToFind.add(cube);
			for(Cube otherCube : piece.cubes)
			{
				List<Cube> cubeFound = new ArrayList<Cube>();
				for(Cube cube : cubeToFind)
				{
					if( cube.equals(otherCube) )
					{
						cubeFound.add(cube);
					}
				}
				cubeToFind.removeAll(cubeFound);
			}
			return cubeToFind.size() == 0;
		}
		return super.equals(obj);
	}
	
	@Override
	public String toString() {
		StringBuffer buffer = new StringBuffer(nom);
		buffer.append("=[");
		for(Cube cube : cubes)
		{
			buffer.append(cube);
		}
		buffer.append("]");
		return buffer.toString();
	}

	/**
	 * @return Le nom de la piece
	 */
	public String getNom() {
		return nom;
	}
	
	@Override
	public Object clone() throws CloneNotSupportedException {
		Cube[] cubesClone = new Cube[cubes.length];
		for(int i = 0; i < cubesClone.length; i++)
		{
			cubesClone[i] = (Cube) cubes[i].clone();
		}
		return new Piece(nom, cubesClone); 
	}

	/**
	 * Dplace la pice  cette position
	 * @param cube La position  atteindre
	 */
	public void moveTo(Cube cube) {
		String toString = null;
		if( logger.isDebugEnabled() )
			toString = toString();
		Piece cible = new Piece("cible", new Cube[]{cube});
		translation(Axe.Abscisse, cible.distanceMoy(this, Axe.Abscisse));
		translation(Axe.Ordonnee, cible.distanceMoy(this, Axe.Ordonnee));
		translation(Axe.Profondeur, cible.distanceMoy(this, Axe.Profondeur));
		if( logger.isDebugEnabled() )
			logger.debug("Dplacement de la piece " + toString + " vers le cube " + cube + " = " + this);
	}
	
	/**
	 * Ajoute un listener
	 * @param propertyChangeListener le listener  ajouter
	 */
	public void addPropertyChangeListener(PropertyChangeListener propertyChangeListener)
	{
		propertyChangeSupport.addPropertyChangeListener(propertyChangeListener);
	}
	
	/**
	 * Supprime un listener
	 * @param propertyChangeListener le listener  ajouter
	 */
	public void removePropertyChangeListener(PropertyChangeListener propertyChangeListener)
	{
		propertyChangeSupport.removePropertyChangeListener(propertyChangeListener);
	}
	
	/**
	 * Retourne les cubes
	 * @return
	 */
	public Cube[] getCubes() {
		return cubes;
	}
}
